



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
euphonic-dispersion --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
euphonic-dos --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
euphonic-optimise-eta --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
euphonic-show-sampling --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
euphonic-intensity-map --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
